/*----------------------------------------------------------------------

 	splitHex

	Split an Intel hex file into two Intel hex files for even bytes
	and odd bytes      

	Written by Mike Douglas
	
	 Ver     Date             Desc
	----   ----------    ------------------------------------------------ 
	 1.0   10/05/2024		Initial version
	 1.1	 10/08/2024		Zero-offset and halve the output addresses

------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define HEX_LINE_LEN 16				/* data bytes per hex output line */
#define FILENAME_LEN 100
#define INBUF_LEN 280
#define MEM_SIZE 64*1024

typedef unsigned char byte;
void writeHex (FILE *, byte *, int, int);

/* main */

int main (int argc, char *argv[])
{
	char	fileName[FILENAME_LEN];
	char	evenExt[16], oddExt[16], *extPtr;
	char	inBuf[INBUF_LEN], dataStr[INBUF_LEN];
	char	*bufPtr;
	byte	*binImage;
	FILE	*inFile, *outFile;
	int	byteCount, dataAddr, recType, dataByte, checkSum, i;
	int	lowAddr = MEM_SIZE-1;
	int	highAddr = 0;

/* Validate command line parameter */

	fprintf(stdout,"\nsplitHex v1.1\n\n");
	if (argc != 2) {
		fprintf(stderr, "Usage: splitHex fname[.ext]\n\n"
							 "Default extension is \"hex\"\n"
							 "Output files are fname-even.hex and fname-odd.hex\n");
		exit(1);
	}

/* Add .hex to file name if no extension provided. Use the same case as
      the user typed for Linux, etc. */

	strcpy(fileName, argv[1]);
	if ((extPtr = strrchr(fileName,'.')) == NULL) {
		extPtr = strrchr(fileName, '\0');			/* get pointer to end of fname */
		if (islower(*fileName)) {
			strcat(fileName, ".hex");
			strcpy(evenExt, "-even.hex");
			strcpy(oddExt, "-odd.hex");
		}
		else {
			strcat(fileName, ".HEX");
			strcpy(evenExt, "-EVEN.HEX");
			strcpy(oddExt, "-ODD.HEX");
		}
	}
	
	if ((inFile = fopen(fileName, "r")) == NULL) {
		fprintf(stderr,"Can't open file %s\n", fileName);
		exit(1);
	}

	if ((binImage = malloc(MEM_SIZE)) == NULL) {
		fprintf(stderr,"Can't allocate image memory\n");
		fclose(inFile);
		exit(1);
	}
	memset(binImage, 0, MEM_SIZE);

/* Read the hex file and put binary data into the memory image array */

	do {
		fgets(inBuf, INBUF_LEN, inFile);
		i = strlen(inBuf)-1;
		if (inBuf[i] == '\n')
			inBuf[i] = '\0';

		sscanf(inBuf, ":%2x%4x%2x%s", &byteCount, &dataAddr, &recType, dataStr);
		checkSum = byteCount + (dataAddr >> 8) + (dataAddr & 0xff) + recType;
		bufPtr = dataStr;

		if ((recType == 0) && (byteCount > 0) && (dataAddr+byteCount <= MEM_SIZE)) {
			if (dataAddr + byteCount > highAddr)
				highAddr = dataAddr + byteCount;
			if (dataAddr < lowAddr)
				lowAddr = dataAddr;
					
			do {
				sscanf(bufPtr, "%2x", &dataByte);
				bufPtr += 2;
				binImage[dataAddr++] = dataByte;
				checkSum += dataByte;
			} while (--byteCount != 0);

			sscanf(bufPtr, "%2x", &dataByte);			/* checksum byte */
			if (0 != ((checkSum+dataByte) & 0xff)) {
				fprintf(stderr,"Checksum error\n  %s\n", inBuf);
				exit(1);
			}
		} 
	} while (!feof(inFile));

	fclose(inFile);

/* Write first hex file  */

	if ((lowAddr & 1) == 0)
		strcpy(extPtr, evenExt);
	else
		strcpy(extPtr, oddExt);

	if ((outFile = fopen(fileName, "w")) == NULL) {
		fprintf(stderr,"Can't open file %s\n", fileName);
		exit(1);
	}
	writeHex(outFile, binImage, lowAddr, highAddr);
	fprintf(stderr,"File %s has been written\n", fileName);

/* Write second hex file  */

	if ((lowAddr & 1) == 0)
		strcpy(extPtr, oddExt);
	else
		strcpy(extPtr, evenExt);

	if ((outFile = fopen(fileName, "w")) == NULL) {
		fprintf(stderr,"Can't open file %s\n", fileName);
		exit(1);
	}
	writeHex(outFile, binImage, lowAddr+1, highAddr);
	fprintf(stderr,"File %s has been written\n", fileName);
}

/*----------------------------------------------------------------------
  Function writeHex
 		Write every other byte of binImage from startAddr to endAddr
 		to outFile in Intel hex format
------------------------------------------------------------------------*/
void writeHex (FILE *outFile, byte *binImage, int startAddr, int endAddr)
{
	int	byteCount, bytesLeft, dataAddr, checkSum, lineAddr, i;

	bytesLeft = (1 + endAddr - startAddr) / 2;		/* round up */
	dataAddr = startAddr;

	do {
		byteCount = bytesLeft > HEX_LINE_LEN ? HEX_LINE_LEN : bytesLeft;
		lineAddr = (dataAddr-startAddr)/2;
		fprintf(outFile, ":%02X%04X00", byteCount, lineAddr);
		checkSum = byteCount + (lineAddr >> 8) + (lineAddr & 0xff);
	
		for (i=0; i<byteCount; i++) {
			fprintf(outFile,"%02X", binImage[dataAddr]);
			checkSum += binImage[dataAddr];
			dataAddr += 2;
		}
		fprintf(outFile, "%02X\n", -checkSum & 0xff);
		bytesLeft -= byteCount;
	} while (bytesLeft > 0);

	fprintf(outFile, ":00000001FF\n");
	fclose(outFile);
}
